/*

Texture routines.

Greg Turk, December 1996.

*/

#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <math.h>
#include <strings.h>
#include "matlib.h"


/******************************************************************************
Get header information (first few lines) from a file.

Entry:
  fd     - file descriptor
  nfeeds - number of lines to get

Exit:
  returns pointer to header info
******************************************************************************/

static char* get_header(int fd, int nfeeds)
{
  char *str;
  int index = 0;
  int linefeeds = 0;
  int done = 0;

  str = (char *) malloc (100);

  while (!done) {
    read(fd, &str[index], 1);
    if (str[index] == '\n')
      linefeeds++;
    if (linefeeds == nfeeds || index >= 100)
      done = 1;
    index++;
  }
  str[index] = '\0';

  if (index >= 100) {
    fprintf (stderr, "Header bad in file.\n");
    exit (-1);
  }

  return (str);
}


/******************************************************************************
Load a texture into memory from a .ppm file.

Entry:
  filename - name of .ppm file to read

Exit:
  returns pointer to texture, or NULL if error occured
******************************************************************************/

Texture *load_texture (char *filename)
{
  char name[80];
  char *str;
  int img_size;
  int fd;
  Texture *texture;

  strcpy (name, filename);
  if (strlen (name) < 4 || strcmp (name + strlen (name) - 4, ".ppm") != 0)
    strcat (name, ".ppm");

  fd = open (name, O_RDONLY);

  if (fd < 0) {
    fprintf(stderr, "Unable to open '%s'\n", filename);
    return (NULL);
  }

  texture = (Texture *) malloc (sizeof (Texture));

  str = get_header(fd, 3);
  sscanf (str, "P6\n%d %d\n255\n", &texture->xsize, &texture->ysize);
  img_size = 3 * texture->xsize * texture->ysize * sizeof(unsigned char);

  texture->data = (char *) malloc (sizeof (char) * img_size);

  if (texture->data == NULL) {
    fprintf(stderr, "Unable to allocate %d bytes for input image\n", img_size);
    return (NULL);
  }

  if (read(fd, texture->data, img_size) != img_size) {
    fprintf(stderr, "Unable to read %d bytes from input image\n", img_size);
    return (NULL);
  }

  close (fd);

  return (texture);
}


/******************************************************************************
Free the memory associated with a texture.
******************************************************************************/

void free_texture (Texture *texture)
{
  free (texture->data);
  free (texture);
}


/******************************************************************************
Get a texture sample based on texture coordinates.

Entry:
  s,t - texture coordinates
  ds_dx,ds_dy - partial derrivatives of texture coordinate s in x and y
  dt_dx,dt_dy - partial derrivatives t

Exit:
  r,g,b - filtered values of the texture at the given position
******************************************************************************/

void get_texture_sample (
  Texture *texture,
  float s,
  float t,
/*  float ds_dx,
  float ds_dy,
  float dt_dx,
  float dt_dy,*/
  float *r,
  float *g,
  float *b
)
{
  float ss,tt;
  int is,it;
  float sfrac,tfrac;
  unsigned char rr,gg,bb;
  int i00,i01,i10,i11;
  float r00,r01,r10,r11;
  float g00,g01,g10,g11;
  float b00,b01,b10,b11;
  float r0,r1;
  float g0,g1;
  float b0,b1;
  unsigned char *d = texture->data;

  /* clamp the texture coordinates to [0,1] */
  if (s < 0) s = 0;
  if (s > 1) s = 1;
  if (t < 0) t = 0;
  if (t > 1) t = 1;



  /* need to flip top-to-bottom */
  t = 1 - t;

  /* compute pixel positions and fractional values */

  if (s == 1) {
    is = texture->xsize-2;
    sfrac = 1;
  }
  else {
    ss = s * (texture->xsize-1);
    is = (int) ss;
    sfrac = ss - is;
  }

  if (t == 1) {
    it = texture->ysize-2;
    tfrac = 1;
  }
  else {

    tt = t * (texture->ysize-1);
    it = (int) tt;
    tfrac = tt - it;
  }

  /* indices of four adjacent pixels-- our sample is between these pixels */

  i00 = 3 * (is     + it     * texture->xsize);
  i01 = 3 * (is     + (it+1) * texture->xsize);
  i10 = 3 * ((is+1) + it     * texture->xsize);
  i11 = 3 * ((is+1) + (it+1) * texture->xsize);




#if 1

  /* shorthands for r,g,b components of the four adjacent pixels */

  r00 = d[i00];
  r01 = d[i01];
  r10 = d[i10];
  r11 = d[i11];

  g00 = d[i00+1];
  g01 = d[i01+1];
  g10 = d[i10+1];
  g11 = d[i11+1];

  b00 = d[i00+2];
  b01 = d[i01+2];
  b10 = d[i10+2];
  b11 = d[i11+2];

  /* perform bi-linear interpolation */

  /* first, interpolate in s direction */

  r0 = r00 + sfrac * (r10 - r00);
  r1 = r01 + sfrac * (r11 - r01);

  g0 = g00 + sfrac * (g10 - g00);
  g1 = g01 + sfrac * (g11 - g01);

  b0 = b00 + sfrac * (b10 - b00);
  b1 = b01 + sfrac * (b11 - b01);

  /* now interpolate in t direction */

  rr = r0 + tfrac * (r1 - r0);
  gg = g0 + tfrac * (g1 - g0);
  bb = b0 + tfrac * (b1 - b0);
#endif

#if 0

  /* point sampling */

  rr = d[i00];
  gg = d[i00+1];
  bb = d[i00+2];
#endif
*r = 0;
*g = 0;
*b = 0;
  *r = 1.0* rr / 255.0;
  *g = 1.0* gg / 255.0;
  *b = 1.0* bb / 255.0;

}


